#------------------------------------------------------------------------------
#
#	Makefile for MPW (Macintosh Programmers Workshop by Apple)
#
#   File:       BlitPixie.make
#   Project:    BlitPixie
#
#	Written by Anders F Bjrklund <afb@algonet.se>
#	
#------------------------------------------------------------------------------

CC68K     		= {C}		# Apple MrC
CCPPC    		= {PPCC}	# Symantec SC
LINK68K 		= Lib
LINKPPC 		= PPCLink

LibName68K		= BlitPixieLib.o
LibNamePPC		= BlitPixieLib.x

Options         = -d USE_ASSEMBLY=0 -d USE_GENERIC_C=0
Warnings        =

ASM68K     		= Asm
ASMPPC    		= PPCAsm

ASM68KOptions   =
ASMPPCOptions	= -w 3200 # avoid warning: "you used r0 instead of 0" (I need to, for old CW)
                          # NOTE: the above gives a warning with PPCAsm < 1.3 <sigh>
                          
#------------------------------------------------------------------------------

Headers      	= -i "::Headers:"

Sources     	=	 "::Sources:"

AsmSources      =    "::Assembler:"

Output      	=    "::Libraries:"


obj68K         	=     :Objects68K:
objPPC	    	=     :ObjectsPPC:

IncludesFolders	=	 {Headers}

SymOptions		=	 -sym off		# turn this on to debug with SADE/R2Db
									# or PowerMac Debugger

Debug68KOptions	=	 -opt speed		# turn this off to build debug 68K code
DebugPPCOptions	=	 -opt speed		# turn this off to build debug PowerPC code

CCOptions       =    {IncludesFolders} {Options} {Warnings} {SymOptions}

CC68KOptions    =	 {CCOptions} {Debug68KOptions} -model far
CCPPCOptions	=	 {CCOptions} {DebugPPCOptions}

LibOptions      =	 {SymOptions} 
LibOptions68K   =	 {LibOptions} 
LibOptionsPPC	=	 {LibOptions} -xm l

Lib68K			=	 {Output}{LibName68K} 
LibPPC  		=	 {Output}{LibNamePPC} 

#------------------------------------------------------------------------------
# These are the objects that we want to include in the library.
#-------------------------------------------------------------------------------

LibObjects68K		=	
		{obj68K}BlitPixieAllBit.a.o 
		{obj68K}BlitPixieClear.a.o 
		{obj68K}BlitPixieCodeCompiler.c.o 
		{obj68K}BlitPixieColorkey.c.o 
		{obj68K}BlitPixieDoubled.c.o 
		{obj68K}BlitPixieDoubled.a.o 
		{obj68K}BlitPixieDoubleRects.a.o 
		{obj68K}BlitPixieFlip.c.o 
		{obj68K}BlitPixieFlip.a.o 
		{obj68K}BlitPixieMask.c.o 
		{obj68K}BlitPixieMask.a.o 
		{obj68K}BlitPixieMaskCollision.a.o 
		{obj68K}BlitPixieRect.a.o 
		{obj68K}BlitPixieRLE.c.o 
		{obj68K}BlitPixieRLE.a.o 
		{obj68K}BlitPixieRLECollision.c.o 
		{obj68K}BlitPixieRLECompiler.c.o 
		{obj68K}BlitPixieRotoZoomer.c.o 
		{obj68K}BlitPixieScale.c.o 
		{obj68K}BlitPixieUtils.c.o 
		{obj68K}BlitPixieUtils.a.o 
		{obj68K}BlitPixieBlitClear.a.o 
		{obj68K}BlitPixieBlitErase.a.o 
		{obj68K}BlitPixieDirtyRects.c.o 
		{obj68K}BlitPixieDirtyRects.a.o 
		#
LibObjectsPPC	=	
		{objPPC}BlitPixieClear.s.x 
		{objPPC}BlitPixieColorkey.c.x 
		{objPPC}BlitPixieDoubled.c.x 
		{objPPC}BlitPixieDoubled.s.x 
		{objPPC}BlitPixieDoubleRects.s.x 
		{objPPC}BlitPixieFlip.c.x 
		{objPPC}BlitPixieFlip.s.x 
		{objPPC}BlitPixieMask.c.x 
		{objPPC}BlitPixieMask.s.x 
		{objPPC}BlitPixieMaskCollision.s.x 
		{objPPC}BlitPixieRect.s.x 
		{objPPC}BlitPixieRLE.c.x 
		{objPPC}BlitPixieRLE.s.x 
		{objPPC}BlitPixieRLECollision.c.x 
		{objPPC}BlitPixieRLECompiler.c.x 
		{objPPC}BlitPixieRotoZoomer.c.x 
		{objPPC}BlitPixieScale.c.x 
		{objPPC}BlitPixieUtils.c.x 
		{objPPC}BlitPixieUtils.s.x 
		{objPPC}BlitPixieBlitClear.s.x 
		{objPPC}BlitPixieBlitErase.s.x 
		{objPPC}BlitPixieDirtyRects.c.x 
		{objPPC}BlitPixieDirtyRects.s.x 
		#

#------------------------------------------------------------------------------
# These are modified default build rules.  
#------------------------------------------------------------------------------

{obj68K}			{Sources}
{objPPC}			{Sources}

.c.o				.c
	Echo "# Compiling {Default}.c using {CC68K}"
	{CC68K} {CC68KOptions} {DepDir}{Default}.c -o {TargDir}{Default}.c.o

.a.o				.c
	Echo "# Compiling {Default}.a using {ASM68K}"
	{ASM68K} {ASM68KOptions} {AsmSources}"68K:"{Default}.a -o {TargDir}{Default}.a.o

.c.x				.c
	Echo "# Compiling {Default}.c using {CCPPC}"
	{CCPPC} {CCPPCOptions} {DepDir}{Default}.c -o {TargDir}{Default}.c.x

.s.x				.c
	Echo "# Compiling {Default}.s using {ASMPPC}"
	{ASMPPC} {ASMPPCOptions} {AsmSources}"PPC:"{Default}.s -o {TargDir}{Default}.s.x

#------------------------------------------------------------------------------
# These are the targets.
#------------------------------------------------------------------------------

BlitPixie			 All
68K.Debug			 68K
PPC.Debug			 PPC
Debug				 All
68K.Final			 68K
PPC.Final			 PPC
Final				 All

All					 68K PPC
68K					 {Lib68K}
PPC					 {LibPPC}

Clean				
	Echo "# Removing objects"
	Delete -i {LibObjects68K} {LibObjectsPPC}
	Delete -y {obj68K} {objPPC}

Clean-All			 Clean
	Echo "# Removing binaries"
	Delete -i {Lib68K} {LibPPC}

#----------------------------------------------

{Lib68K}			 {LibObjects68K}
	Echo "# Linking 680X0 library {LibName68K}"
	{LINK68K} {LibOptions68K} {LibObjects68K} -o {Targ}

{LibPPC}			 {LibObjectsPPC}
	Echo "# Linking PowerPC library {LibNamePPC}"
	{LINKPPC} {LibOptionsPPC} {LibObjectsPPC} -o {Targ}

